#!/usr/bin/env python
from __future__ import print_function
from __future__ import division

import os
import os.path
import subprocess as sp
import re
import pickle

import pandas as pd
pd.set_option('display.max_columns', None)
pd.set_option('display.max_rows', None)        

# configuration values
PATH = os.getenv('NUPACKHOME', default="/Users/nporubsk/nupack/ninja") + "/bin"

class Analysis(object):
    """a translator class between a set of named strands and the analysis"""
    
    def __init__(self, prefix, strands, maxsize, 
                concentrations, complexes_args="", concentrations_args=""):
        super(Analysis, self).__init__()
        self.prefix = prefix
        self.strand_concentrations = concentrations
        self.strands = strands
        self.complexes = Complexes(prefix, strands, maxsize, complexes_args)
        self.concentrations = None
        self.complexes_args = complexes_args
        self.concentrations_args = concentrations_args
        self.translation_location = self.complexes.translation_location
        self.pickle_file = self.prefix + ".pickle"
        self.python_script = self.prefix + ".py"
        self.qsub_script = self.prefix + "_run.sh"
        
        self.pandas = None
        
    def make_qsub_script(self):
        f = open(self.qsub_script, 'w')
        
        cwd = os.getcwd()
        header = "\n".join(["#!/bin/bash",
                "#PBS -S /bin/bash",
                "#PBS -V",
                "#PBS -d {}".format(cwd)])
        
        f.write(header + "\n")
        f.write("cd {}\n".format(cwd))
        f.write("python {}".format(self.python_script))
        
        os.chmod(self.qsub_script, 0o755)
        
        return self.qsub_script
    
    
    def make_python_script(self):
        script = "\n".join([
        "#!/usr/bin/env python",
        "import pickle",
        "x = pickle.load(open(\"{}\", \'r\'))".format(self.pickle_file),
        "x.run()",
        "x.make_readable_output()",
        "x.pickle()",
        "x.remove_intermediate_output()",
        "",
        ])
        
        f = open(self.python_script, "w")
        f.write(script)
        
        os.chmod(self.python_script, 0o755)
        
        return self.python_script
        
    def remove_python_script(self):
        if os.path.isfile(self.python_script):
            os.remove(self.python_script)

        
    def pickle(self):
        pickle.dump(self, open(self.pickle_file, 'w'))
        return self.pickle_file
    
    def remove_pickle(self):
        if os.path.isfile(self.pickle_file):
            os.remove(self.pickle_file)
    
        
    def submit_job(self):
        self.pickle()
        self.make_python_script()
        scriptname = self.make_qsub_script()
        sp.call(["qsub", scriptname])
        
            
    def run(self):
        self.complexes.run()
        
        self.concentrations = Concentrations(self.prefix, 
                self.strand_concentrations, 
                self.complexes.complex_ids,
                self.concentrations_args)
        self.concentrations.run()
        
        
    def make_readable_output(self):
        for x in [self.complexes, self.concentrations]:
            x.translate_output()
            
            
    def make_pandas(self):
        
        a = self.complexes.make_pandas()
        b = self.concentrations.make_pandas()
        self.pandas = a.merge(b, on="Complex Name")
        return self.pandas
            
    
    def remove_intermediate_output(self):
        for x in [self.complexes, self.concentrations]:
            x.remove_base_output()
            


class Complexes(object):
    """
    wrapper for producing, reading, and modifying the complexes
    files to be more human readable.
    """
    
    section_marker = r"% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %"
    _complex_name_regex = re.compile("% composition([0-9]+)-ordering([0-9]+)")
    _command = PATH + '/' + "complexes -ordered -pairs -mfe -quiet {} {}"
    _filenames = {
            "strands": "{}.in",
            "main": "{}.ocx",
            "key": "{}.ocx-key",
            "mfe": "{}.ocx-mfe",
            "pair probabilities": "{}.ocx-ppairs",
            "pair expectations": "{}.ocx-epairs",
            }
    
    
    def __init__(self, prefix, strands, maxsize, additional_args="",):
        super(Complexes, self).__init__()
        self.prefix = prefix
        self.strands = strands
        self.num_strands = len(strands)
        self.maxsize = maxsize
        self.command = self._command.format(additional_args, prefix)
        
        self.strand_ids = {}
        self.complex_ids = {}
        self.complex_sequences = {}
        
        self.filenames = {}
        for x in self._filenames.keys():
            self.filenames[x] = self._filenames[x].format(prefix)
        self.translation_location = self.prefix + "-translated"
                    
        
    def run(self):
        self.make_input()
        sp.call(self.command.split())
    
        
    def make_input(self):
        f = open(self.filenames["strands"], 'w')
        f.write(str(self.num_strands) + "\n")
        for (s, n) in zip(self.strands.keys(), range(1, self.num_strands + 1)):
            f.write(self.strands[s] + "\n")
            self.strand_ids[n] = s
        f.write(str(self.maxsize) + "\n")     
    
        
    def translate_output(self):
        if not os.path.exists(self.translation_location):
            os.mkdir(self.translation_location)
        
        self.parse_complexes()
        self._translate_string_named()
        self._translate_ocx()
    
    
    def make_pandas(self):
        if not os.path.exists(self.translation_location):
            print("Cannot use make_pandas() without translate_output() first")
            return
        
        data = {
                "Complex Name": [],
                "Sequences": [],
                "Free Energy": [],
                "MFE": [],
                "MFE Structure": [],
                "Pair Probabilities": [],
                }
        mfe = self._dict_mfe()
        energy = self._dict_free_energy()
        probs = self._dict_pair_probabilities()
        
        for x in self.complex_sequences.keys():
            data["Complex Name"].append(x)
            data["Sequences"].append(self.complex_sequences[x])
            data["MFE"].append(mfe[x]["MFE"])
            data["MFE Structure"].append(mfe[x]["MFE Structure"])
            data["Free Energy"].append(energy[x])
            data["Pair Probabilities"].append(probs[x])
        
        self.pandas = pd.DataFrame.from_dict(data, orient="columns")
        
        return self.pandas
        
    
    def _dict_mfe(self):
        mfes = {}
        filename = self.translation_location + "/mfe" 
        
        with open(filename, 'r') as f:
            complex_name = ""
            counter = -2
            for line in f:
                x  = line.strip()
                
                if x == self.section_marker: 
                    counter = -1
                elif counter == -1 and '%' in x and x != self.section_marker:
                    complex_name = " ".join(x.split()[1:])
                    counter = 0
                if counter is 0:
                    counter += 1
                elif counter is 1:
                    counter += 1
                    mfes[complex_name] = {"MFE": float(line)}
                elif counter is 2:
                    mfes[complex_name]["MFE Structure"] = line.strip()
                    counter = -1
                
        return mfes
        
        
    def _dict_free_energy(self):
        energy = {}
        filename = self.translation_location + "/free_energies"
        
        with open(filename, 'r') as f:
            for line in f:
                if line[0] is not "%" and len(line) > 0:
                    name, en = [x.strip() for x in line.split(':')]
                    energy[name] = float(en)
                    
        return energy
        
        
    def _dict_pair_probabilities(self):
        probs = {}
        filename = self.translation_location + "/pair_probabilities"
        
        with open(filename, 'r') as f:
            counter = -2
            complex_name = ""
            for line in f:
                x  = line.strip()
                
                if x == self.section_marker: 
                    counter = -1
                elif counter == -1 and '%' in x and x != self.section_marker:
                    complex_name = " ".join(x.split()[1:])
                    counter = 0
                elif counter == 0:
                    length = int(x)
                    probs[complex_name] = []
                    counter = 1
                elif counter == 1:
                    nums = x.split()
                    id = [int(nums[0]), int(nums[1])]
                    id = sorted(id)
                    prob = float(nums[-1])
                    
                    if id[-1] <= length:
                        probs[complex_name].append((id[0], id[1], prob))
                    else:
                        probs[complex_name].append((id[0], id[0], prob))
                                    
        return probs   
    
    
    def line_to_name(self, line):
        match = self._complex_name_regex.match(line)
        if match:
            return self.complex_ids[(int(match.group(1)), int(match.group(2)))]
        else:
            return False
    
                        
    def _translate_string_named(self):
        for x in ['mfe','pair probabilities','pair expectations',]:
            infile = open(self.filenames[x], 'r')
            outfile = open(self.translation_location + '/' + "_".join(x.split()), 'w')
            
            for line in infile:
                complex_name = self.line_to_name(line)
                if complex_name:
                    outfile.write("% {}\n".format(complex_name))
                else:
                    outfile.write(line)
    
                    
    def _translate_ocx(self):
        infile = open(self.filenames['main'])
        outfile = open(self.translation_location + '/' + "free_energies", 'w')
        
        for line in infile:
            if line[0] is '%':
                outfile.write(line)
            else:
                nums = line.split()
                id = (int(nums[0]), int(nums[1]))
                outfile.write("{}: {}\n".format(self.complex_ids[id], nums[-1]))
    
        
    def parse_complexes(self):
        if not os.path.isfile(self.filenames['key']):
            print("must run first to generate output.")
            return
        
        f = open(self.filenames['key'])
        for line in f:
            if line[0] is not '%' and len(line) > 0:
                nums = [int(x) for x in line.split()]
                strand_names = [self.strand_ids[x] for x in nums[2:]]
                complex_name = "-".join(strand_names)
                self.complex_ids[(nums[0], nums[1])] = complex_name
                self.complex_sequences[complex_name] = \
                        [self.strands[strand] for strand in strand_names]
                                
                        
    def remove_base_output(self):
        for x in self.filenames.values():
            if os.path.isfile(x):
                os.remove(x)
            # if x != self.filenames["strands"]:
                # os.remove(x)
    


class Concentrations(object):
    """
    wrapper for producing, reading, and modifying the concentrations
    files to be more human readable.
    """
    
    _command = PATH + '/' + "concentrations -ordered -quiet {} {}"
    _filenames = {
            "complexes output": "{}.ocx",
            "strand concentrations": "{}.con",
            "complex concentrations": "{}.eq"
            }
    
    def __init__(self, prefix, concentrations, id_map, additional_args=""):
        super(Concentrations, self).__init__()
        self.prefix = prefix
        self.concentrations = concentrations
        self.id_map = id_map
        self.command = self._command.format(additional_args, prefix)
        
        self.filenames = {}
        for x in self._filenames.keys():
            self.filenames[x] = self._filenames[x].format(prefix)
        self.translation_location = self.prefix + "-translated"
        
    
    def run(self):
        if not os.path.isfile(self.filenames["complexes output"]):
            print("must run corresponding Complexes first to generate output.")
            return
        
        self.make_input()
        sp.call(self.command.split())
    
        
    def make_input(self):
        f = open(self.filenames["strand concentrations"], 'w')
        for c in self.concentrations:
            f.write(str(c) + "\n")
    
            
    def translate_output(self):
        infile = open(self.filenames['complex concentrations'])
        outfile = open(self.translation_location + '/' + "concentrations", 'w')
        
        for line in infile:
            if line[0] == '%':
                outfile.write(line)
            else:
                nums = line.split()
                id = (int(nums[0]), int(nums[1]))
                outfile.write("{}: {}\n".format(self.id_map[id], nums[-1]))

    
    def make_pandas(self):
        if not os.path.exists(self.translation_location + "/concentrations"):
            print("Cannot use make_pandas() without translate_output() first")
            return
        
        eq = self._dict_eq()
        data = {"Complex Name": [], "Concentration": []}
        for name, conc in eq.items():
            data["Complex Name"].append(name)
            data["Concentration"].append(conc)
                
        self.pandas = pd.DataFrame.from_dict(data, orient="columns")
        
        return self.pandas

    
    def _dict_eq(self):
        eq = {}
        filename = self.translation_location + "/concentrations"
        
        with open(filename, 'r') as f:
            for line in f:
                if line[0] != "%" and len(line) > 0:
                    name, conc = [x.strip() for x in line.split(':')]
                    eq[name] = float(conc)
        
        return eq    
    
    def remove_base_output(self):
        for x in self.filenames.values():
            if os.path.isfile(x):
                os.remove(x)
        # os.remove(self.filenames["complex concentrations"])


if __name__ == '__main__':
    print("NOTE: this file is a set of convenience wrappers for" + 
        " NUPACK and has no executable script code.")
